import sys
import os
from .dir_utils import mkdir_p

IRS_READER_ROOT = os.path.abspath(os.path.dirname(__file__))

# This is the URL to amazon's bucket, could use another synced to it
IRS_XML_HTTP_BASE = "https://s3.amazonaws.com/irs-form-990"

# The directory we put files in while we're processing them
WORKING_DIRECTORY = (os.path.join(IRS_READER_ROOT, "XML"))
# It can be hard to locate this.
IRSX_SETTINGS_LOCATION = (os.path.join(IRS_READER_ROOT, "settings.py"))
# Helpful to keep these around for lookup purposes
INDEX_DIRECTORY = (os.path.join(IRS_READER_ROOT, "CSV"))

KNOWN_SCHEDULES = [
                "IRS990", "IRS990EZ", "IRS990PF", "IRS990ScheduleA",
                "IRS990ScheduleB", "IRS990ScheduleC", "IRS990ScheduleD",
                "IRS990ScheduleE", "IRS990ScheduleF", "IRS990ScheduleG",
                "IRS990ScheduleH", "IRS990ScheduleI", "IRS990ScheduleJ",
                "IRS990ScheduleK", "IRS990ScheduleL", "IRS990ScheduleM",
                "IRS990ScheduleN", "IRS990ScheduleO", "IRS990ScheduleR",
                "ReturnHeader990x"
]

# these should get pushed to metadata directory
ALLOWED_VERSIONSTRINGS = [
    '2013v3.0', '2013v3.1', '2013v4.0', '2014v5.0', '2014v6.0',
    '2015v2.0', '2015v2.1', '2015v3.0', '2016v3.0', '2016v3.1',
    '2017v2.0', '2017v2.1', '2017v2.2', '2017v2.3',
]

CSV_ALLOWED_VERSIONSTRINGS = ALLOWED_VERSIONSTRINGS + [
    '2010v3.2', '2010v3.4', '2010v3.6', '2010v3.7', '2011v1.2', '2011v1.3',
    '2011v1.4', '2011v1.5', '2012v2.0', '2012v2.1', '2012v2.2', '2012v2.3', 
    '2012v3.0'
]

METADATA_DIRECTORY = (os.path.join(IRS_READER_ROOT, "metadata"))

KEYERROR_LOG = os.path.join(IRS_READER_ROOT, "keyerrors.log")
LOG_KEY = 'xml'

mkdir_p([WORKING_DIRECTORY, INDEX_DIRECTORY])

try:
    from .local_settings import *
except ImportError:
    pass


import csv
from operator import itemgetter


CFILE = '/Users/jfenton/github-whitelabel/concordance/irs-efile-master-concordance-file/efiler_master_concordance.csv'
#VARFILE = '/Users/jfenton/github-whitelabel/irsreader/990-xml-reader/irs_reader/metadata/variables.csv'            
VARFILE = '/Users/jfenton/github-whitelabel/irsreader/990-xml-reader/irs_reader/metadata/variables_uniquefied.csv'            

OUTFILE = 'tobemerged.csv'

def get_year_version_numeric(versionstring):
    #print("VS %s" % versionstring)
    if versionstring:
        try:
            last_version = versionstring.strip(";").replace("V","v").split(";")[-1]
            last_version_year = int(last_version.split("v")[0])
            last_version = float(last_version.split("v")[1])

            return(last_version_year, last_version)
        except: 
            return(0,0.0)    
    else:
        return(0,0.0)

def read_concordance():
    """ returns an array of rows """
    reader = csv.DictReader(open(CFILE))
    xpath_hash = {}
    var_hash = {}
    for row in reader:
        (last_version_year, last_version) = get_year_version_numeric(row['version'])
        if last_version_year < 2009:
            continue

        try:
            xpath_hash[row['xpath']] = xpath_hash[row['xpath']] + [row]
        except KeyError:
            xpath_hash[row['xpath']] = [row]
        try:
            var_hash[row['variable_name']] = var_hash[row['variable_name']] + [row]
        except KeyError:
            var_hash[row['variable_name']] = [row]
    return {'xpath_hash':xpath_hash, 'var_hash':var_hash}

def get_varname_from_var_row(row):
    varname = row['db_table'] + '_' + row['db_name']
    return varname

def read_variables():
    reader = csv.DictReader(open(VARFILE))
    xpath_hash = {}
    var_hash = {}

    for row in reader:
        try:
            xpath_hash[row['xpath']]
        except KeyError:
            xpath_hash[row['xpath']] = [row]

        varname = get_varname_from_var_row(row)
        try:
            var_hash[varname] = var_hash[varname] + [row]
        except KeyError:
            var_hash[varname] = [row]
    return {'xpath_hash':xpath_hash, 'var_hash':var_hash}

def transform_from_jf_to_con(var):
    if var.startswith("/ReturnHeader/"):
        return "/Return" + var
    else:
        return "/Return/ReturnData" + var
    ReturnHeader



if __name__ == '__main__':

    cvars = read_concordance()
    c_xpath_hash = cvars['xpath_hash']
    c_var_hash = cvars['var_hash']

    jvars = read_variables()
    j_xpath_hash = jvars['xpath_hash']
    j_var_hash = jvars['var_hash']

    headers = ['variable_name_original', 'merge_into_variable_name','original_versions', 'merge_into_versions', 'original_xpath','merge_into_xpath']
    outfile = open(OUTFILE, 'wb')
    dw = csv.DictWriter(outfile, fieldnames=headers, extrasaction='ignore')
    dw.writeheader()


    def check_overlap(versions1, versions2):
        version_dict = {}
        for version in versions1.split(";"):
            if version:
                version_dict[version] = 1

        for version in versions2.split(";"):
            if version:
                try:
                    version_dict[version]
                    print("**version overlap %s" % version)
                    return True
                except KeyError:
                    pass
        return False


    def check_related_paths(jvar_xpath, xpath, dw):
        msg = ""
        index = 0

        msg += "\nCheck related %s " % jvar_xpath
        xpaths_from_jf = j_xpath_hash.get(jvar_xpath)
        assert len(xpaths_from_jf) == 1
        #xpaths_from_concordance = c_xpath_hash.get(xpath)
        #msg += "\nxpaths from ccd: %s" % xpaths_from_concordance

        ## get the var from each xpath
        thisvar = get_varname_from_var_row(xpaths_from_jf[0])
        jvars = j_var_hash.get(thisvar)


            
        msg += "\nfor jfvar '%s' found %s vars %s" % (thisvar, len(jvars), ["%s %s" % (i['xpath'], i['versions']) for i in jvars])

        ## check that each of the vars has the same concordance var name:
        this_related_cvars = set()
        for jvar in jvars:
            thisjfvar = get_varname_from_var_row(xpaths_from_jf[0])
            this_c_xpaths = c_xpath_hash.get(transform_from_jf_to_con(jvar['xpath']))
            #msg += "\nFound %s for %s" % ([i['variable_name'] for i in this_c_xpaths], jvar['xpath'])
            for this_c_xpath in this_c_xpaths:
                this_related_cvars.add( this_c_xpath['variable_name'] )


        if len(this_related_cvars) > 1:
            #print("\n\n\t ^^^ \nProblem with %s - %s" % (jvar_xpath, this_related_cvars))
            #print(msg)
            cvar_array = list(this_related_cvars)
            #print("cvar array length: %s -> %s" % (len(cvar_array), cvar_array))
            fullcvars = []
            for cvar in cvar_array:
                fullcvar = c_var_hash.get(cvar)[0]
                if fullcvar['version']:
                    (last_version_year, last_version) = get_year_version_numeric(fullcvar['version'])
                    fullcvar['last_version_year'] = last_version_year
                    fullcvar['last_version'] = last_version
                else:
                    fullcvar['last_version'] = 0
                    fullcvar['last_version_year'] = 0

                fullcvars.append(fullcvar)


            fullcvars_sorted = sorted(fullcvars, key=itemgetter('last_version_year', 'last_version'))
            #   headers = ['variable_name_original', 'original_versions','original_xpath', 'merge_into_variable_name', 'merge_into_versions', 'merge_into_xpath']

            versions = fullcvar['version']

            if not check_overlap(fullcvars_sorted[0]['version'], fullcvars_sorted[1]['version']):
            


                index += 1
                row = {
                     'variable_name_original':fullcvars_sorted[0]['variable_name'],
                     'original_versions':fullcvars_sorted[0]['version'],
                     'original_xpath':fullcvars_sorted[0]['xpath'],
                     'merge_into_variable_name':fullcvars_sorted[1]['variable_name'],
                     'merge_into_versions':fullcvars_sorted[1]['version'],
                     'merge_into_xpath':fullcvars_sorted[1]['xpath']
                }


                dw.writerow(row)

                return 1



        # the concordance var repping all these:
        #cvar_name = this_related_cvars.pop()
        #print("jfvar %s from %s contained in %s" % (thisjfvar, jvar['xpath'], cvar_name))
        return 0



    ## Start by going through the jvars one at a time:
    c_keyerror_count = 0
    prob = 0
    for jvar_xpath in j_xpath_hash.keys():
        xpath = transform_from_jf_to_con(jvar_xpath)

        # ignore ones that end in Grp this is a bug
        if xpath.endswith("Grp"):
            continue

        try:
            c_xpath_hash[xpath]
            result = check_related_paths(jvar_xpath, xpath, dw)
            #print("Result is %s" % result)
        except KeyError:
            #print ("KeyError %s" % xpath)
            c_keyerror_count += 1
    print("\n\nTotal keyerrors in concordance: %s" % c_keyerror_count)
    print("Total problem vars in concordance: %s" % prob)





"""
Hand curated list of types that occur in IRS fillings, and their representations
in django and sqlalchemy. The latter is pretty rough.

Defaults to a text field when definitions are missing.

In general the IRS adapts type definitions to shoehorn in old data,
so using current definitions is usually good enough for older stuff... to a point. 

This list is tied to 2013 forwards, maybe this should be namespaced or linked somehow.

USAmountType allows 15 digit ints, so should probably be mapped to biginteger or dealt with.
"""


# A char field longer than MAX_CHAR_FIELD_SIZE will be a text field. 
# Best setting may be db dependent?
MAX_CHAR_FIELD_SIZE = 200 

# based on 2016v3.0, in *general* fields get longer over time, not shorter, but look for conversions from free text to enum. 

var_types = {
    'USAmountType':{'type':'Integer', 'length':15},
    'BooleanType':{'type':'Char', 'length':5}, # http://www.datypic.com/sc/xsd/t-xsd_boolean.html, legal values = [0,1,'true', 'false']
    'USAmountNNType':{'type':'Integer', 'length':15},
    'CheckboxType':{'type':'Char', 'length':1}, # legal values = ['X'] 
    'StreetAddressType':{'type':'Char', 'length':35},
    'LineExplanationType':{'type':'Char', 'length':100},
    'IntegerNNType':{'type':'Integer', 'length':15},  # max value is unclear, see xsd:Integer vs xsd:int 
    'BusinessNameLine2Type':{'type':'Char', 'length':75},
    'BusinessNameLine1Type':{'type':'Char', 'length':75},
    'RatioType':{'type':'Decimal', 'totalDigits':6, 'fractionDigits':5},
    'StateType':{'type':'Char', 'length':2}, # need list key to translate back ? 
    'CountryType':{'type':'Char', 'length':2},
    'ShortExplanationType':{'type':'Text', 'length':1000},
    'CityType':{'type':'Char', 'length':22},
    'ZIPCodeType':{'type':'Char', 'length':15}, # "ZIP Code - 5 digits plus optional 4 or 7 digits"
    'PersonNameType':{'type':'Char', 'length':35},
    'ExplanationType':{'type':'Text', 'length':9000},
    'YearType':{'type':'Integer', 'length':4},
    'EINType':{'type':'Char', 'length':9},
    'DateType':{'type':'Char', 'length':31}, # unclear http://www.datypic.com/sc/xsd/s-datatypes.xsd.html
    'ShortDescriptionType':{'type':'Char', 'length':20},
    'CountType':{'type':'Integer', 'length':7}, # max length is 6
    'Count2Type':{'type':'Integer', 'length':7}, # max length is 6
    'PhoneNumberType':{'type':'Char', 'length':10},
    'IRS990PFPartVDistriRatioType':{'type':'Decimal', 'totalDigits':12, 'fractionDigits':6}, # was 9, manual fix
    'LargeRatioType':{'type':'Decimal', 'totalDigits':22, 'fractionDigits':12},
    'DecimalNNType':{'type':'Decimal', 'totalDigits':22, 'fractionDigits':2}, # dunno upper bound
    'EFINType':{'type':'Char', 'length':6}, # Type for Electronic Filing Identification No. - 6 digits
    'PINType':{'type':'Char', 'length':5}, # Type for Practitioner PIN, Self-Select PIN and Third Party Designee PIN
    'EmailAddressType':{'type':'Char', 'length':75},
    'SoftwareVersionType':{'type':'Char', 'length':20},
    'TimeType':{'type':'Char', 'length':15}, #  Should be no more than 9 chars, but... [0-9]{2}:[0-9]{2}:[0-9]{2}
    'CUSIPNumberType':{'type':'Char', 'length':9},
    'SSNType':{'type':'Char', 'length':12}, # should be 9 but needs to fit "XXX-XX-XXXX" which has 11
    'DeviceIdType':{'type':'Char', 'length':40},
    'BusinessNameControlType':{'type':'Char', 'length':7}, # max is 4: ([A-Z0-9\-]|&#x26;){1,4}
    'PersonTitleType':{'type':'Char', 'length':35}, 
    'OriginatorType':{'type':'Char', 'length':15},
    'TimestampType':{'type':'Char', 'length':63}, # not sure
    'ISPType':{'type':'Char', 'length':6}, # Type for Intermediate Service Provider No. - 6 uppercase alphanumeric characters
    'PTINType':{'type':'Char', 'length':9}, # Type for Preparer Personal Identification No. - P followed by 8 digits
    'USAmountNegType':{'type':'Integer', 'length':15},
    'IPv6Type':{'type':'Char', 'length':31}, # 1
    'SoftwareIdType':{'type':'Char', 'length':8}, #The Software ID - 8 digits
    'IPv4Type':{'type':'Char', 'length':31},
    'InCareOfNameType':{'type':'Char', 'length':35},
    'TimezoneType':{'type':'Char', 'length':31}
}

def get_django_type(vartype):

    try:
        thisvar = var_types[vartype]
    except KeyError:
        return None

    if (thisvar['type'] =='Integer'): 
        if thisvar['length'] < 10:
            return "IntegerField(null=True, blank=True)" # 32 bit: Values from -2147483648 to 2147483647 
        else:
            return "BigIntegerField(null=True, blank=True)" # 64 bit: from: -9223372036854775808 to 9223372036854775807

    elif (thisvar['type']=='Decimal'):
        return "DecimalField(null=True, blank=True, max_digits=%s, decimal_places=%s)" % (thisvar['totalDigits'], thisvar['fractionDigits'])

    elif (thisvar['type']=='Char'):
        if thisvar['length'] <= MAX_CHAR_FIELD_SIZE: 
            return "CharField(null=True, blank=True, max_length=%s)" % thisvar['length']
        else:
            return "TextField(null=True, blank=True)"

    elif (thisvar['type']=='Text'):
        return "TextField(null=True, blank=True)"

    else:
        print("** No match for %s " % thisvar)
        return "TextField(null=True, blank=True)"

def get_sqlalchemy_type(vartype):

    """ This is really rough, not tested, may change """ 

    try:
        thisvar = var_types[vartype]
    except KeyError:
        return None

    if (thisvar['type'] =='Integer'): 
        if thisvar['length'] < 10: # 64 bit should be forced to BigInteger
            return "Integer" 
        else:
            return "BigInteger" 

    elif (thisvar['type']=='Decimal'):
        return "Numeric(precision=%s)" % (thisvar['totalDigits']) 

    elif (thisvar['type']=='Char'):
        if thisvar['length'] <= MAX_CHAR_FIELD_SIZE: 
            return "String(length=%s)" % thisvar['length']
        else:
            return "Text"

    elif (thisvar['type']=='Text'):
        return "Text"

    else:
        print("** No match for %s " % thisvar)
        return "Text"

if __name__ == "__main__":
    for key in var_types.keys():
        print("key %s resolve to '%s' and '%s'" % (key, get_django_type(key), get_sqlalchemy_type(key) ) ) 




""" alterative to xml_runner that uses unparsed sked_reader """
import csv
import os

from .filing import Filing
from .standardizer import Standardizer, Documentizer, VersionDocumentizer
from .text_format_utils import debracket
#from .sked_dict_reader import SkedDictReader
from .unparsed_sked_reader import UnparsedSkedReader as SkedDictReader
# from .log_utils import configure_logging
from .type_utils import listType

from .settings import WORKING_DIRECTORY, ALLOWED_VERSIONSTRINGS, METADATA_DIRECTORY

# local override 
# 201303199349310500 is version 2012v2.1
ALLOWED_VERSIONSTRINGS = ALLOWED_VERSIONSTRINGS + ['2010v3.2', '2010v3.4', '2010v3.6', '2010v3.7', '2011v1.2', '2011v1.3', '2011v1.4', '2011v1.5', '2012v2.0', '2012v2.1', '2012v2.2', '2012v2.3', '2012v3.0']
ALLOWED_VERSIONSTRINGS = ALLOWED_VERSIONSTRINGS +  ['2009v1.0', ' 2009v1.1', ' 2009v1.2', ' 2009v1.3', ' 2009v1.4', ' 2009v1.7']

class NDC_Documentizer(object):
    """
    Returns data from NDC concordance to add for hack before datathon
    """

    def __init__(self):
        self.ndc_lines = {}
        self._make_ndc_lines()


    def _make_ndc_lines(self):
        filepath = os.path.join(METADATA_DIRECTORY, 'efiler_master_concordance.csv')
        with open(filepath, 'r') as reader_fh:
            reader = csv.DictReader(reader_fh)
            for row in reader:
                self.ndc_lines[row['xpath']] = row

    def get_ndc_line(self, xpath):
        if xpath.startswith("/IRS990"):
            xpath = "/Return/ReturnData" + xpath
        else:
            xpath = "/Return" + xpath

        try:
            return self.ndc_lines[xpath]
        except KeyError:
            return None



class UnparsedXMLRunner(object):
    """ Load a Standardizer just once while running multiple filings
        Return Filing objects with results, keyerrors set
    """
    def __init__(self, documentation=False, standardizer=None):
        self.documentation = documentation

        if documentation:
            # we need a new standardizer, ignore one if passed in
            self.standardizer = Documentizer()
        else:
            if standardizer:
                self.standardizer = standardizer
            else:
                self.standardizer = Standardizer()
        self.group_dicts = self.standardizer.get_groups()
        self.whole_filing_data = []
        self.filing_keyerr_data = []

    def get_standardizer(self):
        return self.standardizer

    def _run_schedule_k(self, sked, object_id, sked_dict, path_root, ein):
        assert sked == 'IRS990ScheduleK'
        if type(sked_dict) == listType:
            for individual_sked in sked_dict:
                doc_id = individual_sked['@documentId']
                reader = SkedDictReader(
                    self.standardizer,
                    self.group_dicts,
                    object_id,
                    ein,
                    documentId=doc_id,
                    documentation=self.documentation
                )

                result = reader.parse(individual_sked, parent_path=path_root)
                self.whole_filing_data.append({
                    'schedule_name': sked,
                    'groups': result['groups'],
                    'schedule_parts': result['schedule_parts'],
                    'csv_line_array':result['csv_line_array']
                })
        else:
            reader = SkedDictReader(
                self.standardizer,
                self.group_dicts,
                object_id,
                ein,
                documentation=self.documentation)

            result = reader.parse(sked_dict, parent_path=path_root)
            self.whole_filing_data.append({
                'schedule_name': sked,
                'groups': result['groups'],
                'schedule_parts': result['schedule_parts'],
                'csv_line_array':result['csv_line_array']

            })

    def _run_schedule(self, sked, object_id, sked_dict, ein):
        print("_run_schedule %s" % sked)
        path_root = "/" + sked
        # Only sked K (bonds) is allowed to repeat
        if sked == 'IRS990ScheduleK':
            self._run_schedule_k(sked, object_id, sked_dict, path_root, ein)

        else:
            reader = SkedDictReader(
                self.standardizer,
                self.group_dicts,
                object_id,
                ein,
                documentation=self.documentation
            )
            if sked == 'ReturnHeader990x':
                path_root = "/ReturnHeader"
            result = reader.parse(sked_dict, parent_path=path_root)
            self.whole_filing_data.append({
                'schedule_name': sked,
                'groups': result['groups'],
                'schedule_parts': result['schedule_parts'],
                'csv_line_array':result['csv_line_array']

            })

            if len(result['group_keyerrors']) > 0 or len(result['keyerrors'])> 0:
                self.filing_keyerr_data.append({
                    'schedule_name': sked,
                    'group_keyerrors':result['group_keyerrors'],
                    'keyerrors':result['keyerrors']
                })

    def run_filing(self, object_id, verbose=False):
        self.whole_filing_data = []
        self.filing_keyerr_data = []
        this_filing = Filing(object_id)
        this_filing.process(verbose=verbose)
        this_version = this_filing.get_version()
        if verbose:
            print("Filing %s is version %s" % (object_id, this_version))
        if this_version in ALLOWED_VERSIONSTRINGS:
            this_version = this_filing.get_version()
            schedules = this_filing.list_schedules()
            ein = this_filing.get_ein()
            self.whole_filing_data = []
            for sked in schedules:
                sked_dict = this_filing.get_schedule(sked)
                self._run_schedule(sked, object_id, sked_dict, ein)

            this_filing.set_result(self.whole_filing_data)
            this_filing.set_keyerrors(self.filing_keyerr_data)
            if verbose:
                if len(self.filing_keyerr_data)>0:
                    pass
                    #print("In %s keyerrors: %s" % (object_id, self.filing_keyerr_data))
                else:
                    print("No keyerrors found")
            return this_filing
        else:
            return this_filing

    def run_from_filing_obj(self, this_filing, verbose=False):  
        """
         Run from a pre-created filing object.
        """
        self.whole_filing_data = []
        self.filing_keyerr_data = []
        this_filing.process(verbose=verbose)
        object_id = this_filing.get_object_id()
        this_version = this_filing.get_version()
        if this_version in ALLOWED_VERSIONSTRINGS:
            this_version = this_filing.get_version()
            schedules = this_filing.list_schedules()
            ein = this_filing.get_ein()
            for sked in schedules:
                sked_dict = this_filing.get_schedule(sked)
                self._run_schedule(sked, object_id, sked_dict, ein)
            this_filing.set_result(self.whole_filing_data)
            this_filing.set_keyerrors(self.filing_keyerr_data)
            return this_filing
        else:
            return this_filing


    def run_sked(self, object_id, sked, verbose=False):
        """
        sked is the proper name of the schedule:
        IRS990, IRS990EZ, IRS990PF, IRS990ScheduleA, etc.
        """
        self.whole_filing_data = []
        self.filing_keyerr_data = []
        this_filing = Filing(object_id)
        this_filing.process(verbose=verbose)
        this_version = this_filing.get_version()
        if this_version in ALLOWED_VERSIONSTRINGS:
            this_version = this_filing.get_version()
            ein = this_filing.get_ein()
            sked_dict = this_filing.get_schedule(sked)
            self._run_schedule(sked, object_id, sked_dict, ein)

            this_filing.set_result(self.whole_filing_data)
            this_filing.set_csv_result(self.csv_line_array)
            this_filing.set_keyerrors(self.filing_keyerr_data)
            return this_filing
        else:
            return this_filing

if __name__ == '__main__':

    #test_set = ['201511319349301766', '201403179349305520', '201542189349301214', '201530449349302683', '201503159349304000', '201530419349300113', '201613209349309896', '201603549349300640', '201503289349300110', '201623209349310262', '201613509349300606', '201543589349300104', '201633519349301053', '201633579349300718', '201613199349306136', '201533139349300208', '201603149349302590', '201532299349304633', '201603179349200000', '201643199349302309', '201623169349100822', '201543169349101804', '201613139349100776', '201630679349300208', '201621349349101032']
    
    test_set = ['201511319349301766', '201403179349305520', '201542189349301214', '201530449349302683', '201503159349304000', '201530419349300113', '201613209349309896', '201603549349300640', '201503289349300110', '201623209349310262', '201613509349300606', '201543589349300104', '201633519349301053', '201633579349300718', '201613199349306136', '201533139349300208', '201603149349302590', '201532299349304633', '201603179349200000', '201643199349302309', '201623169349100822', '201543169349101804', '201613139349100776', '201630679349300208', '201621349349101032', '201123159349301107', '201113349349300206', '201143189349307054', '201103139349300915', '201133199349100833', '201123199349203627', '201533209349304768', '201533179349307818', '201513089349301656', '201533189349300608', '201103159349304200', '201123129349301977', '201513359349100306', '201630529349200103', '201602259349201355', '201601369349200905']
    #test_set = ['201613509349300606']
    vd = VersionDocumentizer()
    #object_id = '201303199349310500' # v2012v2.1
    #object_id = '201113139349301336' # 2010v3.2
    #object_id = '201533089349301428'  #multiple schedule K's
    #test_set = ['201123159349301107', '201113349349300206', '201143189349307054', '201103139349300915', '201133199349100833', '201123199349203627', '201533209349304768', '201533179349307818', '201513089349301656', '201533189349300608', '201103159349304200', '201123129349301977', '201513359349100306', '201630529349200103', '201602259349201355', '201601369349200905']
    xml_runner = UnparsedXMLRunner(documentation=True)
    standardizer = xml_runner.get_standardizer()
    nd = NDC_Documentizer()

    for object_id in test_set:
        parsed_filing = xml_runner.run_filing(
                    object_id,
                    verbose=True
                )

        #fieldnames = ['location_code', 'value', 'is_error', 'error_description', 'description', 'line_number', 'xpath', 'variable_name', 'in_group', 'group_name', 'group_index']
        fieldnames = ['value', 'description', 'line_number', 'xpath', 'in_group', 'group_index']

        outfilename = '/Users/jfenton/github-whitelabel/validata/createdcsvs/' + object_id + ".csv"
        outfile = open(outfilename, 'w') # 'wb' python 2?
        writer = csv.DictWriter(
            outfile,
            fieldnames=fieldnames,
            delimiter=',',
            quotechar='"',
            lineterminator='\n',
            quoting=csv.QUOTE_MINIMAL,
            extrasaction='ignore'
        )
        writer.writeheader()

        results = parsed_filing.get_result()

        for result in results:
            for this_result in result['csv_line_array']:


                raw_line_num = vd.get_line_number(this_result['xpath'], parsed_filing.get_version())
                this_result['line_number'] =  debracket(raw_line_num)

                raw_description = vd.get_description(this_result['xpath'], parsed_filing.get_version())
                this_result['description'] =  debracket(raw_description)

                nd_result = nd.get_ndc_line(this_result['xpath'])
                if nd_result:
                    this_result['location_code'] = nd_result['location_code']
                    this_result['variable_name'] = nd_result['variable_name']

                if this_result['group_index']:
                    if this_result['group_index'] > 0:
                        this_result['is_error'] = 'DO NOT VALIDATE'
                        this_result['error_description'] = 'DO NOT VALIDATE'

                writer.writerow(this_result)




""" Helper to check if a given version is in a version range """

import re

version_re = re.compile(r'20\d\dv\d\.\d\Z')
strict=True

def validate_version(version):
    if not version_re.match(version):
        raise ValueError("Illegal version string '%s'" % version)
    return True

def check_version(version_start, version_end, version_test):

    if strict:
        [validate_version(i) for i in [version_start, version_end, version_test]]


if __name__ == '__main__':
    check_version('2013v3.0', '2014v2.3', '2014v1.0')

    


""" 
Regenerate the models and documentation from the .csv files.
Use the standardizer and documentizer to feed data. 
""" 
import os
import re
import unidecode

from .standardizer import Documentizer
from .settings import KNOWN_SCHEDULES, METADATA_DIRECTORY
from .irs_type_utils import get_django_type, get_sqlalchemy_type

FILENAME = "django_models_auto.py"
OUTFILE = os.path.join(METADATA_DIRECTORY, FILENAME)
SOFT_TAB = '    '
BRACKET_RE = re.compile(r'\[.*?\]')


def debracket(string):
    """ Eliminate the bracketed var names in doc, line strings """
    result = re.sub(BRACKET_RE, ';', string)
    result = unidecode.unidecode(result)
    return result

def most_recent(semicolon_delimited_string):
    result = semicolon_delimited_string.split(";")[-1]
    return result

class ModelRegenerator(object):

    def __init__(self, django=True):

        self.run_django = django
        self.run_sqlalchemy = not django

        self.documentizer = Documentizer()

        self.sked_parts = self.documentizer.get_schedule_parts()
        self.variables = self.documentizer.get_variables()
        self.outfile = open(OUTFILE, 'w')
        self.table_name_dict = {}
        for var in self.variables.keys():
            vardata = self.variables[var]
            #print(vardata['db_table'])
            try:
                self.table_name_dict[vardata['db_table']].append(vardata)
            except KeyError:
                print("Table: %s" % vardata['db_table'])
                self.table_name_dict[vardata['db_table']] = [vardata]


    def write_model_top(self, sked_name, full_name, parent_sked_name, repeating_group_part=None, repeating_group_description=None):

        if self.run_django:

            result = "\n#######\n#\n# %s - %s\n" % (parent_sked_name, full_name)
            if repeating_group_description:
                result += "# Description: %s\n" % (repeating_group_description)
            if repeating_group_part:
                result += "# A repeating structure from %s\n" % (repeating_group_part)
            result += "#\n#######\n"
            ## write the start of the first group:
            result += "\nclass %s(models.Model):\n" % sked_name
            result +=  SOFT_TAB +  "object_id = models.CharField(max_length=31, blank=True, null=True, help_text=\"unique xml return id\")\n"
            result +=  SOFT_TAB +  "ein = models.CharField(max_length=15, blank=True, null=True, help_text=\"filer EIN\")\n"
            if parent_sked_name=='IRS990ScheduleK':
                # It's not clear what the max length is; Return.xsd is unclear
                result +=  SOFT_TAB +  "documentId = models.TextField(blank=True, null=True, help_text=\"documentID attribute\")"


            return result

        elif self.run_sqlalchemy:

            result = "\n#######\n#\n# %s - %s\n" % (parent_sked_name, full_name)
            if repeating_group_description:
                result += "# Description: %s\n" % (repeating_group_description)
            if repeating_group_part:
                result += "# A repeating structure from %s\n" % (repeating_group_part)
            result += "#\n#######\n"
            ## write the start of the first group:
            result += "\nclass %s(Base):\n%s__tablename__='%s'\n" % (sked_name,SOFT_TAB, sked_name)
            result +=  SOFT_TAB +  "object_id = Column(String(31))\n"
            result +=  SOFT_TAB +  "ein = Column(String(15))\n"
            if parent_sked_name=='IRS990ScheduleK':
                result +=  SOFT_TAB +  "documentId = Column(String(15))\n"

            result +=  SOFT_TAB +  "id = Column(Integer, primary_key=True)\n" # Add a primary key explicitly

            return result



    def write_top_matter(self):
        if self.run_django:
            self.outfile.write("# -*- coding: latin-1 -*-\n")   # latin1 chars keep getting in from editor        
            self.outfile.write("from django.db import models\n")
        elif self.run_sqlalchemy:
            self.outfile.write("from sqlalchemy import Column, Integer, String, BigInteger, Text, Numeric\n")
            self.outfile.write("from sqlalchemy.ext.declarative import declarative_base\n\n")
            self.outfile.write("Base = declarative_base()\n\n")


    def write_variable(self, variable):
        """
        We fallback to a text field, but we expect the types to be filled in where missing
        """
        #print("write_var %s" % variable)

        if self.run_django:
            variable_output = get_django_type(variable['irs_type'])
            if not variable_output:
                variable_output = "TextField(null=True, blank=True)"
            result =  "\n" + SOFT_TAB + "%s = models.%s" % (variable['db_name'], variable_output)

        elif self.run_sqlalchemy:
            variable_output = get_sqlalchemy_type(variable['irs_type'])
            if not variable_output:
                variable_output = "Text"
            result =  "\n" + SOFT_TAB + "%s = Column(%s)" % (variable['db_name'], variable_output)

        # add newline and documentation regardless of where it's going
        result += "\n" + SOFT_TAB + "#"
        if variable['line_number']:
            result += " Line number: %s " % most_recent(debracket(variable['line_number']))
        if variable['description']:
            result += " Description: %s " % most_recent(debracket(variable['description']))
        result += " xpath: %s \n" % variable['xpath']

        return result


    def regenerate(self):

        self.write_top_matter()

        for sked in KNOWN_SCHEDULES:
            print("Schedule %s" % sked)
            sked_parts = self.documentizer.get_parts_by_sked(sked)
            print("Num sked parts %s " % len(sked_parts))
            for part in sked_parts:
                if part['is_shell']==True:
                    continue
                    print("Skipping shell %s" % part['name'])

                model_top = self.write_model_top(part['parent_sked_part'], part['name'], part['parent_sked'], repeating_group_part=None)
                self.outfile.write(model_top)
                vars = None
                try:
                    vars = self.table_name_dict[part['parent_sked_part']]
                    print("+++ not a keyerror %s" % part['parent_sked_part'])

                except KeyError:
                    print("--- keyerror %s" % part['parent_sked_part'])
                    continue

                for var in vars:

                    self.outfile.write(self.write_variable(var))
                    pass


            groups_in_this_sked = self.documentizer.get_groups_by_sked(sked)
            for group in groups_in_this_sked:
                print("\tgroup:%s" % group['db_name'])

                model_top = self.write_model_top(group['db_name'], group['db_name'], group['parent_sked'], repeating_group_part=group['parent_sked_part'], repeating_group_description=group['description'])
                self.outfile.write(model_top)
                vars = None
                try:
                    vars = self.table_name_dict[group['db_name']]
                except KeyError:
                    print(">>>> keyerror %s" % part['parent_sked_part'])
                    continue
                for var in vars:
                    self.outfile.write(self.write_variable(var))
                    pass

        self.outfile.close()


if __name__ == '__main__':
    R = ModelRegenerator()
    R.regenerate()


import csv

from .standardizer import Documentizer

CFILE = '/Users/jfenton/github-whitelabel/concordance/irs-efile-master-concordance-file/efiler_master_concordance.csv'
            
def read_concordance():
    """ returns an array of rows """

    reader = csv.DictReader(open(CFILE))
    xpath_hash = {}
    var_hash = {}
    for row in reader:
        #print(row)
        try:
            xpath_hash[row['xpath']] = xpath_hash[row['xpath']] + [row]
        except KeyError:
            xpath_hash[row['xpath']] = [row]

        try:
            var_hash[row['variable_name']] = var_hash[row['variable_name']] + [row]
        except KeyError:
            var_hash[row['variable_name']] = [row]

    return {'xpath_hash':xpath_hash, 'var_hash':var_hash}


def get_concordance_var_version(var_array, version):
    for var in var_array:
        if version in var['version'].split(';'):
            print ("Got %s %s" % (var['variable_name'], version))
            return var['variable_name']

    return None


if __name__ == '__main__':

    cvars = read_concordance()
    xpath_hash = cvars['xpath_hash']
    var_hash = cvars['var_hash']

    documentizer = Documentizer(versions=True)
    variables = documentizer.get_variables()
    # for var in variables:
    #     #print(var)
    #     thisvar = documentizer.get_var(var)

    #     this_xpath = thisvar['xpath']
        
    #     if this_xpath.startswith("/IRS990"):
    #         this_xpath = "/Return/ReturnData" + this_xpath
    #     else:
    #         this_xpath = "/Return" + this_xpath

    #     this_jfvar = thisvar['db_table'] + '_' + thisvar['db_name']
    #     modern_concordance_var = None   
    #     try:    
    #         modern_concordance_var = get_concordance_var_version(xpath_hash[this_xpath], '2015v3.0')
        
    #     except KeyError:
    #         print("keyerror for %s" % this_xpath)
    #         continue

    #     if modern_concordance_var:
    #         pass
    #         print("Found %s <-> %s : %s" % (modern_concordance_var, this_xpath, thisvar ))
            
    ## run through variables, find them in concordance, then extend them
    print("Now looking to extend post 2013 vars")
    for var in variables:
        thisvar = documentizer.get_var(var)
        this_jfvar = thisvar['db_table'] + '_' + thisvar['db_name']
        print ("Handling var %s, %s" % (var, this_jfvar))


